"""Scenic world model for the X-Plane flight simulator.

See the `VerifAI distribution`_ for examples of how to use Scenic with X-Plane.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model

.. _VerifAI distribution: https://github.com/BerkeleyLearnVerify/VerifAI
"""
