"""Scenic's core types and associated support code.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   distributions
   external_params
   geometry
   lazy_eval
   object_types
   pruning
   regions
   scenarios
   specifiers
   type_support
   utils
   vectors
   workspaces
"""

