"""Scenic world model for Grand Theft Auto V (GTAV).

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model
   interface
   center_detection
   img_modf
   map
   messages
"""
