"""World model for road intersection scenarios in Webots.

This is a more specialized version of the :doc:`scenic.simulators.webots.road` model which also includes guideway information from the `Intelligent Intersections Toolkit <https://github.com/ucbtrans/intelligent_intersection>`_.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model
   intersection
   interface
"""

